use std::io::{stdin, Read, Stdin};

const BS: u8 = 8;
const ESC: u8 = 27;
const TAB: u8 = b'\t';
const LF: u8 = b'\n';
const CR: u8 = b'\r';
const DEL: u8 = 127;

#[derive(Debug)]
pub enum Event {
	Key(Key, Mod),
	Unsupported(Vec<u8>),
	// may add mouse events at some point
}

#[derive(Debug)]
pub enum Mod {
	None,
	Shift,
	Ctrl,
	Alt,
	/// only works for some specific keys
	ShiftCtrl,
	/// only works for some specific keys
	ShiftAlt,
	/// only works for some specific keys
	CtrlAlt,
	/// only works for some specific keys
	ShiftCtrlAlt,
}

#[non_exhaustive]
#[derive(Debug)]
pub enum Key {
	Char(char),
	Right,
	Left,
	Up,
	Down,
	Home,
	End,
	PageUp,
	PageDown,
	/// also Ctrl+H
	Backspace,
	Delete,
	/// also Ctrl+[
	Escape,
	/// \r or \n or Ctrl+M or Ctrl+J
	Enter,
	/// \t or Ctrl+I
	Tab,
	F(u8),
}

pub struct Events {
	source: Stdin,
	buffer: Vec<u8>,
}

impl Events {
	pub fn new() -> Self {
		Self {
			source: stdin(),
			buffer: Vec::new(),
		}
	}

	fn parse_esc(&mut self) -> Event {
		if self.buffer.is_empty() {
			return Event::Key(Key::Escape, Mod::None);
		}

		let byte = self.buffer[0];
		self.buffer = self.buffer[1..].to_vec();
		match byte {
			ESC => Event::Key(Key::Escape, Mod::None),
			BS => Event::Key(Key::Backspace, Mod::Alt),
			DEL => Event::Key(Key::Delete, Mod::Alt),
			LF | CR => Event::Key(Key::Enter, Mod::Alt),
			TAB => Event::Key(Key::Tab, Mod::Alt),
			c @ 1..=26 => Event::Key(Key::Char((c + 96).into()), Mod::CtrlAlt), // ctrl + letter
			c @ b' '..=b'~' => Event::Key(Key::Char(c.into()), Mod::Alt),       // regular ascii charaters
			_ => Event::Unsupported(vec![ESC, byte]),
		}
	}
}

impl Iterator for Events {
	type Item = Event;

	fn next(&mut self) -> Option<Self::Item> {
		self.source.read_to_end(&mut self.buffer).ok()?;
		if self.buffer.is_empty() {
			return None;
		}

		let byte = self.buffer[0];
		self.buffer = self.buffer[1..].to_vec();
		Some(match byte {
			ESC => self.parse_esc(),
			BS => Event::Key(Key::Backspace, Mod::None),
			DEL => Event::Key(Key::Delete, Mod::None),
			LF | CR => Event::Key(Key::Enter, Mod::None),
			TAB => Event::Key(Key::Tab, Mod::None),
			c @ 1..=26 => Event::Key(Key::Char((c + 96).into()), Mod::Ctrl), // ctrl + letter
			c @ b' '..=b'~' => Event::Key(Key::Char(c.into()), Mod::None),   // regular ascii charaters
			_ => Event::Unsupported(vec![byte]),
		})
	}
}
